@@ -280,6 +280,12 @@ namespace XrdCl
           // so we need to nullify the pointer
           pHandler = 0;
         }
+        else
+        {
+          delete status;
+          delete response;
+          delete hostList;
+        }
         // destroy the object if it is
         DestroyMyself( scopedLock );
       }
@@ -341,6 +347,7 @@ namespace XrdCl
     ResetMonitoringVars();
     DefaultEnv::GetForkHandler()->RegisterFileObject( this );
     DefaultEnv::GetFileTimer()->RegisterFileObject( this );
+    pLFileHandler = new LocalFileHandler();
   }
 
   //------------------------------------------------------------------------
@@ -370,6 +377,7 @@ namespace XrdCl
     ResetMonitoringVars();
     DefaultEnv::GetForkHandler()->RegisterFileObject( this );
     DefaultEnv::GetFileTimer()->RegisterFileObject( this );
+    pLFileHandler = new LocalFileHandler();
   }
 
   //----------------------------------------------------------------------------
@@ -406,6 +414,7 @@ namespace XrdCl
     delete pDataServer;
     delete pLoadBalancer;
     delete [] pFileHandle;
+    delete pLFileHandler;
   }
 
   //----------------------------------------------------------------------------
@@ -441,6 +450,11 @@ namespace XrdCl
 
     if (pFileUrl)
     {
+      if( pUseVirtRedirector && pFileUrl->IsMetalink() )
+      {
+        RedirectorRegistry& registry = RedirectorRegistry::Instance();
+        registry.Release( *pFileUrl );
+      }
       delete pFileUrl;
       pFileUrl = 0;
     }
@@ -486,14 +500,32 @@ namespace XrdCl
 
     pOpenMode  = mode;
     pOpenFlags = flags;
+    OpenHandler *openHandler = new OpenHandler( this, handler );
+
+    bool redirect = pUseVirtRedirector && pFileUrl->IsMetalink();
+
+    //--------------------------------------------------------------------------
+    // If we don't want to redirect and it's a local file simply delgate
+    // to the LocalFileHandler
+    //--------------------------------------------------------------------------
+    if( pFileUrl->IsLocalFile() && !redirect )
+    {
+      Status st = pLFileHandler->Open( pFileUrl->GetURL().c_str(), flags,
+                                             mode, openHandler, timeout );
+      if( !st.IsOK() )
+      {
+        delete openHandler;
+        pStatus    = st;
+        pFileState = Error;
+      }
+
+      return st;
+    }
 
     Message           *msg;
     ClientOpenRequest *req;
-    std::string        path = pFileUrl->GetPathWithParams();
-    if( pUseVirtRedirector && pFileUrl->IsMetalink() )
-      MessageUtils::CreateRequest<VirtualMessage>( msg, req, path.length() );
-    else
-      MessageUtils::CreateRequest( msg, req, path.length() );
+    std::string        path = pFileUrl->GetPathWithFilteredParams();
+    MessageUtils::CreateRequest( msg, req, path.length() );
 
     req->requestid = kXR_open;
     req->mode      = mode;
@@ -502,27 +534,17 @@ namespace XrdCl
     msg->Append( path.c_str(), path.length(), 24 );
 
     XRootDTransport::SetDescription( msg );
-    OpenHandler *openHandler = new OpenHandler( this, handler );
     MessageSendParams params; params.timeout = timeout;
     params.followRedirects = pFollowRedirects;
     MessageUtils::ProcessSendParams( params );
 
-    //--------------------------------------------------------------------------
-    // Register a virtual redirector
-    //--------------------------------------------------------------------------
-    if( pUseVirtRedirector && pFileUrl->IsMetalink() )
-    {
-      RedirectorRegistry& registry = RedirectorRegistry::Instance();
-      XRootDStatus st = registry.Register( *pFileUrl );
-      if( !st.IsOK() ) return st;
-      HostInfo info( url, true );
-      HostList *list = new HostList();
-      list->push_back( info );
-      params.loadBalancer = info;
-      params.hostList     = list;
-    }
-
-    Status st = MessageUtils::SendMessage( *pFileUrl, msg, openHandler, params );
+    Status st;
+    if( redirect )
+      st = MessageUtils::RedirectMessage( *pFileUrl, msg, openHandler,
+                                          params, pLFileHandler );
+    else
+      st = MessageUtils::SendMessage( *pFileUrl, msg, openHandler,
+                                      params, pLFileHandler );
 
     if( !st.IsOK() )
     {
@@ -581,7 +603,12 @@ namespace XrdCl
     params.stateful        = true;
     MessageUtils::ProcessSendParams( params );
 
-    Status st = MessageUtils::SendMessage( *pDataServer, msg, closeHandler, params );
+    Status st;
+    if( pDataServer->IsLocalFile() )
+      st = pLFileHandler->Close( closeHandler, timeout );
+    else
+      st = MessageUtils::SendMessage( *pDataServer, msg, closeHandler,
+                                      params, pLFileHandler );
 
     if( !st.IsOK() )
     {
@@ -649,6 +676,13 @@ namespace XrdCl
 
     XRootDTransport::SetDescription( msg );
     StatefulHandler *stHandler = new StatefulHandler( this, handler, msg, params );
+
+    if( pDataServer->IsLocalFile() )
+    {
+      XRootDStatus st = pLFileHandler->Stat( stHandler, timeout );
+      return ExamineLocalResult( st, msg, stHandler );
+    }
+
     return SendOrQueue( *pDataServer, msg, stHandler, params );
   }
 
@@ -692,6 +726,13 @@ namespace XrdCl
     MessageUtils::ProcessSendParams( params );
 
     StatefulHandler *stHandler = new StatefulHandler( this, handler, msg, params );
+
+    if( pDataServer->IsLocalFile() )
+    {
+      XRootDStatus st = pLFileHandler->Read( offset, size, buffer, stHandler, timeout );
+      return ExamineLocalResult( st, msg, stHandler );
+    }
+
     return SendOrQueue( *pDataServer, msg, stHandler, params );
   }
 
@@ -736,6 +777,13 @@ namespace XrdCl
 
     XRootDTransport::SetDescription( msg );
     StatefulHandler *stHandler = new StatefulHandler( this, handler, msg, params );
+
+    if( pDataServer->IsLocalFile() )
+    {
+      XRootDStatus st = pLFileHandler->Write( offset, size, buffer, stHandler, timeout );
+      return ExamineLocalResult( st, msg, stHandler );
+    }
+
     return SendOrQueue( *pDataServer, msg, stHandler, params );
   }
 
@@ -770,6 +818,13 @@ namespace XrdCl
 
     XRootDTransport::SetDescription( msg );
     StatefulHandler *stHandler = new StatefulHandler( this, handler, msg, params );
+
+    if( pDataServer->IsLocalFile() )
+    {
+      XRootDStatus st = pLFileHandler->Sync( stHandler, timeout );
+      return ExamineLocalResult( st, msg, stHandler );
+    }
+
     return SendOrQueue( *pDataServer, msg, stHandler, params );
   }
 
@@ -806,6 +861,13 @@ namespace XrdCl
 
     XRootDTransport::SetDescription( msg );
     StatefulHandler *stHandler = new StatefulHandler( this, handler, msg, params );
+
+    if( pDataServer->IsLocalFile() )
+    {
+      XRootDStatus st = pLFileHandler->Truncate( size, stHandler, timeout );
+      return ExamineLocalResult( st, msg, stHandler );
+    }
+
     return SendOrQueue( *pDataServer, msg, stHandler, params );
   }
 
@@ -879,6 +941,73 @@ namespace XrdCl
 
     XRootDTransport::SetDescription( msg );
     StatefulHandler *stHandler = new StatefulHandler( this, handler, msg, params );
+
+    if( pDataServer->IsLocalFile() )
+    {
+      XRootDStatus st = pLFileHandler->VectorRead( *list, buffer, stHandler, timeout );
+      return ExamineLocalResult( st, msg, stHandler );
+    }
+
+    return SendOrQueue( *pDataServer, msg, stHandler, params );
+  }
+
+  //------------------------------------------------------------------------
+  // Write scattered buffers in one operation - async
+  //------------------------------------------------------------------------
+  XRootDStatus FileStateHandler::WriteV( uint64_t            offset,
+                                              const struct iovec *iov,
+                                              int                 iovcnt,
+                                              ResponseHandler    *handler,
+                                              uint16_t            timeout )
+  {
+    XrdSysMutexHelper scopedLock( pMutex );
+
+    if( pFileState != Opened && pFileState != Recovering )
+      return XRootDStatus( stError, errInvalidOp );
+
+    Log *log = DefaultEnv::GetLog();
+    log->Debug( FileMsg, "[0x%x@%s] Sending a write command for handle 0x%x to "
+                "%s", this, pFileUrl->GetURL().c_str(),
+                *((uint32_t*)pFileHandle), pDataServer->GetHostId().c_str() );
+
+    Message            *msg;
+    ClientWriteRequest *req;
+    MessageUtils::CreateRequest( msg, req );
+
+    ChunkList *list   = new ChunkList();
+
+    uint32_t size = 0;
+    for( int i = 0; i < iovcnt; ++i )
+    {
+      size += iov[i].iov_len;
+      list->push_back( ChunkInfo( 0, iov[i].iov_len,
+                       (char*)iov[i].iov_base ) );
+    }
+
+    req->requestid  = kXR_write;
+    req->offset     = offset;
+    req->dlen       = size;
+    memcpy( req->fhandle, pFileHandle, 4 );
+
+
+
+    MessageSendParams params;
+    params.timeout         = timeout;
+    params.followRedirects = false;
+    params.stateful        = true;
+    params.chunkList       = list;
+
+    MessageUtils::ProcessSendParams( params );
+
+    XRootDTransport::SetDescription( msg );
+    StatefulHandler *stHandler = new StatefulHandler( this, handler, msg, params );
+
+    if( pDataServer->IsLocalFile() )
+    {
+      XRootDStatus st = pLFileHandler->WriteV( offset, iov, iovcnt, stHandler, timeout );
+      return ExamineLocalResult( st, msg, stHandler );
+    }
+
     return SendOrQueue( *pDataServer, msg, stHandler, params );
   }
 
@@ -918,6 +1047,13 @@ namespace XrdCl
 
     XRootDTransport::SetDescription( msg );
     StatefulHandler *stHandler = new StatefulHandler( this, handler, msg, params );
+
+    if( pDataServer->IsLocalFile() )
+    {
+      XRootDStatus st = pLFileHandler->Fcntl( arg, stHandler, timeout );
+      return ExamineLocalResult( st, msg, stHandler );
+    }
+
     return SendOrQueue( *pDataServer, msg, stHandler, params );
   }
 
@@ -953,6 +1089,12 @@ namespace XrdCl
 
     XRootDTransport::SetDescription( msg );
     StatefulHandler *stHandler = new StatefulHandler( this, handler, msg, params );
+    if( pFileUrl->IsLocalFile() )
+    {
+      XRootDStatus st = pLFileHandler->Visa( stHandler, timeout );
+      return ExamineLocalResult( st, msg, stHandler );
+    }
+
     return SendOrQueue( *pDataServer, msg, stHandler, params );
   }
 
@@ -1436,7 +1578,8 @@ namespace XrdCl
     if( pFileState == Opened )
     {
       msg->SetSessionId( pSessionId );
-      Status st = MessageUtils::SendMessage( *pDataServer, msg, handler, sendParams );
+      Status st = MessageUtils::SendMessage( *pDataServer, msg, handler,
+                                             sendParams, pLFileHandler );
 
       //------------------------------------------------------------------------
       // Invalid session id means that the connection has been broken while we
@@ -1567,7 +1710,8 @@ namespace XrdCl
 
     MessageUtils::ProcessSendParams( params );
 
-    return MessageUtils::SendMessage( *pDataServer, msg, handler, params );
+    return MessageUtils::SendMessage( *pDataServer, msg, handler,
+                                      params, pLFileHandler );
   }
 
   //----------------------------------------------------------------------------
@@ -1599,11 +1743,8 @@ namespace XrdCl
     if( !url.GetPath().empty() )
       u.SetPath( pFileUrl->GetPath() );
 
-    std::string path = u.GetPathWithParams();
-    if( pUseVirtRedirector && pFileUrl->IsMetalink() )
-      MessageUtils::CreateRequest<VirtualMessage>( msg, req, path.length() );
-    else
-      MessageUtils::CreateRequest( msg, req, path.length() );
+    std::string path = u.GetPathWithFilteredParams();
+    MessageUtils::CreateRequest( msg, req, path.length() );
 
     req->requestid = kXR_open;
     req->mode      = pOpenMode;
@@ -1629,19 +1770,20 @@ namespace XrdCl
     MessageUtils::ProcessSendParams( params );
     XRootDTransport::SetDescription( msg );
 
+    Status st;
     //--------------------------------------------------------------------------
     // If a virtual redirector is in use, set it up in send parameters
+    // and redirect the message
     //--------------------------------------------------------------------------
-    if( pUseVirtRedirector && pFileUrl->IsMetalink() )
-    {
-      HostInfo info( *pFileUrl, true );
-      HostList *list = new HostList();
-      list->push_back( info );
-      params.loadBalancer = info;
-      params.hostList     = list;
-    }
-
-    Status st = MessageUtils::SendMessage( url, msg, openHandler, params );
+    if( pUseVirtRedirector && url.IsMetalink() )
+      st = MessageUtils::RedirectMessage( url, msg, openHandler,
+                                          params, pLFileHandler );
+    //--------------------------------------------------------------------------
+    // Otherwise do an ordinary send
+    //--------------------------------------------------------------------------
+    else
+      st = MessageUtils::SendMessage( url, msg, openHandler,
+                                      params, pLFileHandler );
     // if there was a problem destroy the open handler
     if( !st.IsOK() )
     {
@@ -1708,7 +1850,8 @@ namespace XrdCl
       it->request->SetSessionId( pSessionId );
       ReWriteFileHandle( it->request );
       Status st = MessageUtils::SendMessage( *pDataServer, it->request,
-                                             it->handler, it->params );
+                                             it->handler, it->params,
+                                             pLFileHandler );
       if( !st.IsOK() )
         FailMessage( *it, st );
     }
@@ -1787,4 +1930,15 @@ namespace XrdCl
       mon->Event( Monitor::EvClose, &i );
     }
   }
+
+  XRootDStatus FileStateHandler::ExamineLocalResult( XRootDStatus &status,
+                                                     Message *msg,
+                                                     ResponseHandler *handler )
+  {
+    if( status.IsOK() )
+      pInTheFly.insert( msg );
+    else
+      delete handler;
+    return status;
+  }
 }
