@@ -40,6 +40,65 @@
 
 namespace
 {
+
+  //----------------------------------------------------------------------------
+  // Get delimiter for the opaque info
+  //----------------------------------------------------------------------------
+   char GetCgiDelimiter( bool &hasCgi )
+   {
+     if( !hasCgi )
+     {
+       hasCgi = true;
+       return '?';
+     }
+
+     return '&';
+   }
+  //----------------------------------------------------------------------------
+  // Filters out client specific CGI
+  //----------------------------------------------------------------------------
+  std::string FilterXrdClCgi( const std::string &path )
+  {
+    // first check if there's an opaque info at all
+    size_t pos = path.find( '?' );
+    if( pos == std::string::npos )
+      return path;
+
+    std::string filteredPath = path.substr( 0 , pos );
+    std::string cgi = path.substr( pos + 1 );
+
+    bool hasCgi = false;
+    pos = 0;
+    size_t xrdcl = std::string::npos;
+    do
+    {
+      xrdcl = cgi.find( "xrdcl.", pos );
+
+      if( xrdcl == std::string:: npos )
+      {
+        filteredPath += GetCgiDelimiter( hasCgi );
+        filteredPath += cgi.substr( pos );
+        pos = cgi.size();
+      }
+      else
+      {
+        if( xrdcl != pos )
+        {
+          filteredPath += GetCgiDelimiter( hasCgi );
+          filteredPath += cgi.substr( pos, xrdcl - 1 - pos );
+        }
+
+        pos = cgi.find( '&', xrdcl );
+        if( pos != std::string::npos )
+          ++pos;
+      }
+
+    }
+    while( pos < cgi.size() && pos != std::string::npos );
+
+    return filteredPath;
+  }
+
   //----------------------------------------------------------------------------
   //! Wrapper class used to delete FileSystem object
   //----------------------------------------------------------------------------
@@ -407,9 +466,23 @@ namespace
           {
             DirectoryList::ListEntry *entry = *itr;
             StatInfo *info = entry->GetStatInfo();
+            if( !info )
+              throw RecDirLsErr( XRootDStatus( stError, errNotSupported ) );
             std::string path = dirList->GetParentName() + entry->GetName();
 
-            if( info->TestFlags( StatInfo::IsDir ) ) // it is a directory
+            // check the prefix
+            if( path.find( parent ) != 0 )
+              throw RecDirLsErr( XRootDStatus( stError, errInternal ) );
+
+            // add new entry to the result
+            path = path.substr( parent.size() );
+            entry->SetStatInfo( 0 ); // StatInfo is no longer owned by dirList
+            DirectoryList::ListEntry *e =
+                new DirectoryList::ListEntry( entry->GetHostAddress(), path, info );
+            pCtx->dirList->Add( e );
+
+            // if it's a directory do a recursive call
+            if( info->TestFlags( StatInfo::IsDir ) )
             {
               // bump the pending counter
               ++pCtx->pending;
@@ -429,22 +502,10 @@ namespace
                   throw RecDirLsErr( XRootDStatus( stError, errOperationExpired ) );
               }
               // send the request
-              XRootDStatus st = pCtx->fs->DirList( path, flags, handler, timeout );
+              XRootDStatus st = pCtx->fs->DirList( parent + path, flags, handler, timeout );
               if( !st.IsOK() )
                 throw RecDirLsErr( st );
             }
-            else // it's a file
-            {
-              // check the prefix
-              if( path.find( parent ) != 0 )
-                throw RecDirLsErr( XRootDStatus( stError, errDataError ) );
-
-              path = path.substr( parent.size() );
-              entry->SetStatInfo( 0 ); // StatInfo is no longer owned by dirList
-              DirectoryList::ListEntry *e =
-                  new DirectoryList::ListEntry( entry->GetHostAddress(), path, info );
-              pCtx->dirList->Add( e );
-            }
           }
 
           if( pCtx->pending == 0 )
@@ -464,17 +525,14 @@ namespace
           }
         }
 
+        // clean up the context if necessary
+        bool delctx = ( pCtx->pending == 0 );
+        scoped.UnLock();
+        if( delctx )
+          delete pCtx;
         // clean up the arguments
         delete status;
         delete response;
-        // clean up the context if necessary
-        if( pCtx->pending == 0 )
-        {
-          scoped.UnLock();
-          delete pCtx;
-        }
-        else
-          scoped.UnLock();
         // and finally commit suicide
         delete this;
       }
@@ -484,6 +542,127 @@ namespace
 
       RecursiveDirListCtx *pCtx;
   };
+
+  //----------------------------------------------------------------------------
+  // Exception for a merge dirlist handler
+  //----------------------------------------------------------------------------
+  struct MergeDirLsErr
+  {
+      MergeDirLsErr( XrdCl::XRootDStatus *&status, XrdCl::AnyObject *&response ) :
+        status( status ), response( response )
+      {
+        status = 0; response = 0;
+      }
+
+      MergeDirLsErr() :
+        status( new XrdCl::XRootDStatus( XrdCl::stError, XrdCl::errInternal ) ),
+        response( 0 )
+      {
+
+      }
+
+      XrdCl::XRootDStatus *status;
+      XrdCl::AnyObject    *response;
+  };
+
+
+
+  //----------------------------------------------------------------------------
+  // Handle results for a merge dirlist request
+  //----------------------------------------------------------------------------
+  class MergeDirListHandler: public XrdCl::ResponseHandler
+  {
+    public:
+
+      MergeDirListHandler( XrdCl::ResponseHandler *handler ) : pHandler( handler )
+      {
+
+      }
+
+      virtual void HandleResponse( XrdCl::XRootDStatus *status,
+                                   XrdCl::AnyObject    *response )
+      {
+        try
+        {
+          if( status->IsOK() )
+            throw MergeDirLsErr( status, response );
+
+          if( !response )
+            throw MergeDirLsErr();
+
+          XrdCl::DirectoryList *dirlist = 0;
+          response->Get( dirlist );
+
+          if( !dirlist )
+            throw MergeDirLsErr();
+
+          Merge( dirlist );
+          response->Set( dirlist );
+          pHandler->HandleResponse( status, response );
+        }
+        catch( const MergeDirLsErr &err )
+        {
+          delete status; delete response;
+          pHandler->HandleResponse( err.status, err.response );
+        }
+
+        delete this;
+      }
+
+      static void Merge( XrdCl::DirectoryList *&response )
+      {
+        std::set<ListEntry*, less> unique( response->Begin(), response->End() );
+
+        XrdCl::DirectoryList *dirlist = new XrdCl::DirectoryList();
+        dirlist->SetParentName( response->GetParentName() );
+        for( auto itr = unique.begin(); itr != unique.end(); ++itr )
+        {
+          ListEntry *entry = *itr;
+          dirlist->Add( new ListEntry( entry->GetHostAddress(),
+                                       entry->GetName(),
+                                       entry->GetStatInfo() ) );
+          entry->SetStatInfo( 0 );
+        }
+
+        delete response;
+        response = dirlist;
+      }
+
+    private:
+
+      typedef XrdCl::DirectoryList::ListEntry ListEntry;
+
+      struct less
+      {
+        bool operator() (const ListEntry *x, const ListEntry *y) const
+        {
+          if( x->GetName() != y->GetName() )
+            return x->GetName() < y->GetName();
+
+          const XrdCl::StatInfo *xStatInfo = x->GetStatInfo();
+          const XrdCl::StatInfo *yStatInfo = y->GetStatInfo();
+
+          if( xStatInfo == yStatInfo )
+            return false;
+
+          if( xStatInfo == 0 )
+            return true;
+
+          if( yStatInfo == 0 )
+            return false;
+
+          if( xStatInfo->GetSize() != yStatInfo->GetSize() )
+            return xStatInfo->GetSize() < yStatInfo->GetSize();
+
+          if( xStatInfo->GetFlags() != yStatInfo->GetFlags() )
+            return xStatInfo->GetFlags() < yStatInfo->GetFlags();
+
+          return false;
+        }
+      };
+
+      XrdCl::ResponseHandler *pHandler;
+  };
 }
 
 namespace XrdCl
@@ -588,14 +767,16 @@ namespace XrdCl
     if( pPlugIn )
       return pPlugIn->Locate( path, flags, handler, timeout );
 
+    std::string fPath = FilterXrdClCgi( path );
+
     Message             *msg;
     ClientLocateRequest *req;
-    MessageUtils::CreateRequest( msg, req, path.length() );
+    MessageUtils::CreateRequest( msg, req, fPath.length() );
 
     req->requestid = kXR_locate;
     req->options   = flags;
-    req->dlen      = path.length();
-    msg->Append( path.c_str(), path.length(), 24 );
+    req->dlen      = fPath.length();
+    msg->Append( fPath.c_str(), fPath.length(), 24 );
     MessageSendParams params; params.timeout = timeout;
     MessageUtils::ProcessSendParams( params );
 
@@ -661,16 +842,19 @@ namespace XrdCl
     if( pPlugIn )
       return pPlugIn->Mv( source, dest, handler, timeout );
 
+    std::string fSource = FilterXrdClCgi( source );
+    std::string fDest   = FilterXrdClCgi( dest );
+
     Message         *msg;
     ClientMvRequest *req;
-    MessageUtils::CreateRequest( msg, req, source.length()+dest.length()+1 );
+    MessageUtils::CreateRequest( msg, req, fSource.length() + fDest.length()+1 );
 
     req->requestid = kXR_mv;
-    req->dlen      = source.length()+dest.length()+1;
-    req->arg1len   = source.length();
-    msg->Append( source.c_str(), source.length(), 24 );
-    *msg->GetBuffer(24+source.length()) = ' ';
-    msg->Append( dest.c_str(), dest.length(), 25+source.length() );
+    req->dlen      = fSource.length() + fDest.length()+1;
+    req->arg1len   = fSource.length();
+    msg->Append( fSource.c_str(), fSource.length(), 24 );
+    *msg->GetBuffer(24 + fSource.length()) = ' ';
+    msg->Append( fDest.c_str(), fDest.length(), 25 + fSource.length() );
     MessageSendParams params; params.timeout = timeout;
     MessageUtils::ProcessSendParams( params );
 
@@ -747,14 +931,16 @@ namespace XrdCl
     if( pPlugIn )
       return pPlugIn->Truncate( path, size, handler, timeout );
 
+    std::string fPath = FilterXrdClCgi( path );
+
     Message               *msg;
     ClientTruncateRequest *req;
-    MessageUtils::CreateRequest( msg, req, path.length() );
+    MessageUtils::CreateRequest( msg, req, fPath.length() );
 
     req->requestid = kXR_truncate;
     req->offset    = size;
-    req->dlen      = path.length();
-    msg->Append( path.c_str(), path.length(), 24 );
+    req->dlen      = fPath.length();
+    msg->Append( fPath.c_str(), fPath.length(), 24 );
     MessageSendParams params; params.timeout = timeout;
     MessageUtils::ProcessSendParams( params );
     XRootDTransport::SetDescription( msg );
@@ -787,13 +973,15 @@ namespace XrdCl
     if( pPlugIn )
       return pPlugIn->Rm( path, handler, timeout );
 
+    std::string fPath = FilterXrdClCgi( path );
+
     Message         *msg;
     ClientRmRequest *req;
-    MessageUtils::CreateRequest( msg, req, path.length() );
+    MessageUtils::CreateRequest( msg, req, fPath.length() );
 
     req->requestid = kXR_rm;
-    req->dlen      = path.length();
-    msg->Append( path.c_str(), path.length(), 24 );
+    req->dlen      = fPath.length();
+    msg->Append( fPath.c_str(), fPath.length(), 24 );
     MessageSendParams params; params.timeout = timeout;
     MessageUtils::ProcessSendParams( params );
     XRootDTransport::SetDescription( msg );
@@ -827,15 +1015,17 @@ namespace XrdCl
     if( pPlugIn )
       return pPlugIn->MkDir( path, flags, mode, handler, timeout );
 
+    std::string fPath = FilterXrdClCgi( path );
+
     Message            *msg;
     ClientMkdirRequest *req;
-    MessageUtils::CreateRequest( msg, req, path.length() );
+    MessageUtils::CreateRequest( msg, req, fPath.length() );
 
     req->requestid  = kXR_mkdir;
     req->options[0] = flags;
     req->mode       = mode;
-    req->dlen       = path.length();
-    msg->Append( path.c_str(), path.length(), 24 );
+    req->dlen       = fPath.length();
+    msg->Append( fPath.c_str(), fPath.length(), 24 );
     MessageSendParams params; params.timeout = timeout;
     MessageUtils::ProcessSendParams( params );
     XRootDTransport::SetDescription( msg );
@@ -869,13 +1059,15 @@ namespace XrdCl
     if( pPlugIn )
       return pPlugIn->RmDir( path, handler, timeout );
 
+    std::string fPath = FilterXrdClCgi( path );
+
     Message            *msg;
     ClientRmdirRequest *req;
-    MessageUtils::CreateRequest( msg, req, path.length() );
+    MessageUtils::CreateRequest( msg, req, fPath.length() );
 
     req->requestid  = kXR_rmdir;
-    req->dlen       = path.length();
-    msg->Append( path.c_str(), path.length(), 24 );
+    req->dlen       = fPath.length();
+    msg->Append( fPath.c_str(), fPath.length(), 24 );
     MessageSendParams params; params.timeout = timeout;
     MessageUtils::ProcessSendParams( params );
     XRootDTransport::SetDescription( msg );
@@ -908,14 +1100,16 @@ namespace XrdCl
     if( pPlugIn )
       return pPlugIn->ChMod( path, mode, handler, timeout );
 
+    std::string fPath = FilterXrdClCgi( path );
+
     Message            *msg;
     ClientChmodRequest *req;
-    MessageUtils::CreateRequest( msg, req, path.length() );
+    MessageUtils::CreateRequest( msg, req, fPath.length() );
 
     req->requestid  = kXR_chmod;
     req->mode       = mode;
-    req->dlen       = path.length();
-    msg->Append( path.c_str(), path.length(), 24 );
+    req->dlen       = fPath.length();
+    msg->Append( fPath.c_str(), fPath.length(), 24 );
     MessageSendParams params; params.timeout = timeout;
     MessageUtils::ProcessSendParams( params );
     XRootDTransport::SetDescription( msg );
@@ -982,14 +1176,16 @@ namespace XrdCl
     if( pPlugIn )
       return pPlugIn->Stat( path, handler, timeout );
 
+    std::string fPath = FilterXrdClCgi( path );
+
     Message           *msg;
     ClientStatRequest *req;
-    MessageUtils::CreateRequest( msg, req, path.length() );
+    MessageUtils::CreateRequest( msg, req, fPath.length() );
 
     req->requestid  = kXR_stat;
     req->options    = 0;
-    req->dlen       = path.length();
-    msg->Append( path.c_str(), path.length(), 24 );
+    req->dlen       = fPath.length();
+    msg->Append( fPath.c_str(), fPath.length(), 24 );
     MessageSendParams params; params.timeout = timeout;
     MessageUtils::ProcessSendParams( params );
     XRootDTransport::SetDescription( msg );
@@ -1022,14 +1218,16 @@ namespace XrdCl
     if( pPlugIn )
       return pPlugIn->StatVFS( path, handler, timeout );
 
+    std::string fPath = FilterXrdClCgi( path );
+
     Message           *msg;
     ClientStatRequest *req;
-    MessageUtils::CreateRequest( msg, req, path.length() );
+    MessageUtils::CreateRequest( msg, req, fPath.length() );
 
     req->requestid  = kXR_stat;
     req->options    = kXR_vfs;
-    req->dlen       = path.length();
-    msg->Append( path.c_str(), path.length(), 24 );
+    req->dlen       = fPath.length();
+    msg->Append( fPath.c_str(), fPath.length(), 24 );
     MessageSendParams params; params.timeout = timeout;
     MessageUtils::ProcessSendParams( params );
     XRootDTransport::SetDescription( msg );
@@ -1099,20 +1297,26 @@ namespace XrdCl
     if( pPlugIn )
       return pPlugIn->DirList( path, flags, handler, timeout );
 
+    URL url = URL( path );
+    std::string fPath = FilterXrdClCgi( path );
+
     Message           *msg;
     ClientDirlistRequest *req;
-    MessageUtils::CreateRequest( msg, req, path.length() );
+    MessageUtils::CreateRequest( msg, req, fPath.length() );
 
     req->requestid  = kXR_dirlist;
-    req->dlen       = path.length();
+    req->dlen       = fPath.length();
 
     if( ( flags & DirListFlags::Stat ) || ( flags & DirListFlags::Recursive ) )
       req->options[0] = kXR_dstat;
 
     if( flags & DirListFlags::Recursive )
-      handler = new RecursiveDirListHandler( *pUrl, path, flags, handler, timeout );
+      handler = new RecursiveDirListHandler( *pUrl, url.GetPath(), flags, handler, timeout );
 
-    msg->Append( path.c_str(), path.length(), 24 );
+    if( flags & DirListFlags::Merge )
+      handler = new MergeDirListHandler( handler );
+
+    msg->Append( fPath.c_str(), fPath.length(), 24 );
     MessageSendParams params; params.timeout = timeout;
     MessageUtils::ProcessSendParams( params );
     XRootDTransport::SetDescription( msg );
@@ -1191,6 +1395,9 @@ namespace XrdCl
       }
       delete locations;
 
+      if( flags & DirListFlags::Merge )
+        MergeDirListHandler::Merge( response );
+
       if( errors || partial )
       {
         if( errors == numLocations )
@@ -1416,6 +1623,6 @@ namespace XrdCl
 
     params.followRedirects = pFollowRedirects;
 
-    return MessageUtils::SendMessage( *pUrl, msg, handler, params );
+    return MessageUtils::SendMessage( *pUrl, msg, handler, params, 0 );
   }
 }
